/*	TSignal.cpp - Created by Giampiero Caprino

This file is part of Train Director 3

Train Director is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; using exclusively version 2.
It is expressly forbidden the use of higher versions of the GNU
General Public License.

Train Director is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Train Director; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.
*/

#include <stdio.h>
#include <string.h>

#if !defined(__unix__) && !defined(__WXMAC__)
#include <malloc.h>
#endif

#ifdef WIN32
#define strcasecmp stricmp
#endif

#include "Traindir3.h"
#include "itracks.h"
#include "itools.h"
#include "isignals.h"
#include "iswitch.h"
#include "iactions.h"
#include "imovers.h"
#include "TDFile.h"

int	pathIsBusy(Train *tr, Vector *path, int dir);
Vector	*findPath(Track *t, int dir);
Track	*findNextTrack(trkdir direction, int x, int y);

int	trace_script = 0;

extern	int	signals_changed;

static  void    *n_sig_pmap[2];         /* R, G */
static	void	*n_sigx_pmap[2];
static const char *n_sig_xpm[] = {
"7 9 3 1",
"       c #FFFFFFFFFFFF",
".      c #000000000000",
NULL, /*"X      c #0000FFFFFFFF",*/
"       ",
"  ...  ",
" .GGG. ",
" .GGG. ",
" .GGG. ",
"  ...  ",
"   .   ",
"   .   ",
" ..... "};
static const char *n_sigx_xpm[] = {
"7 9 3 1",
"       c #FFFFFFFFFFFF",
".      c #000000000000",
NULL, /*"X      c #0000FFFFFFFF",*/
"       ",
" ..... ",
" .GGG. ",
" .GGG. ",
" .GGG. ",
" ..... ",
"   .   ",
"   .   ",
" ..... "};

static  void    *s_sig_pmap[2];         /* R, G */
static  void    *s_sigx_pmap[2];
static const char *s_sig_xpm[] = {
"7 9 3 1",
"       c #FFFFFFFFFFFF",
".      c #000000000000",
NULL, /*"X      c #0000FFFFFFFF",*/
"       ",
" ..... ",
"   .   ",
"   .   ",
"  ...  ",
" .GGG. ",
" .GGG. ",
" .GGG. ",
"  ...  "};
static const char *s_sigx_xpm[] = {
"7 9 3 1",
"       c #FFFFFFFFFFFF",
".      c #000000000000",
NULL, /*"X      c #0000FFFFFFFF",*/
"       ",
" ..... ",
"   .   ",
"   .   ",
" ..... ",
" .GGG. ",
" .GGG. ",
" .GGG. ",
" ..... "};

static  void    *n_sig2_pmap[4];         /* R, G */
static  void    *n_sig2x_pmap[4];
static const char *n_sig2_xpm[] = {
"7 13 4 1",
"       c #FFFFFFFFFFFF",
".      c #000000000000",
NULL, /*"G      c #0000FFFFFFFF",*/
NULL, /*"X      c #0000FFFFFFFF",*/
"  ...  ",
" .GGG. ",
" .GGG. ",
" .GGG. ",
"  ...  ",
"  ...  ",
" .XXX. ",
" .XXX. ",
" .XXX. ",
"  ...  ",
"   .   ",
"   .   ",
" ..... "};
static const char *n_sig2x_xpm[] = {
"7 13 4 1",
"       c #FFFFFFFFFFFF",
".      c #000000000000",
NULL, /*"G      c #0000FFFFFFFF",*/
NULL, /*"X      c #0000FFFFFFFF",*/
" ..... ",
" .GGG. ",
" .GGG. ",
" .GGG. ",
" ..... ",
" ..... ",
" .XXX. ",
" .XXX. ",
" .XXX. ",
" ..... ",
"   .   ",
"   .   ",
" ..... "};

static  void    *s_sig2_pmap[4];         /* R, G */
static  void    *s_sig2x_pmap[4];         /* R, G */
static const char *s_sig2_xpm[] = {
"7 13 4 1",
"       c #FFFFFFFFFFFF",
".      c #000000000000",
NULL, /*"G      c #0000FFFFFFFF",*/
NULL, /*"X      c #0000FFFFFFFF",*/
" ..... ",
"   .   ",
"   .   ",
"  ...  ",
" .XXX. ",
" .XXX. ",
" .XXX. ",
"  ...  ",
"  ...  ",
" .GGG. ",
" .GGG. ",
" .GGG. ",
"  ...  "};
static const char *s_sig2x_xpm[] = {
"7 13 4 1",
"       c #FFFFFFFFFFFF",
".      c #000000000000",
NULL, /*"G      c #0000FFFFFFFF",*/
NULL, /*"X      c #0000FFFFFFFF",*/
" ..... ",
"   .   ",
"   .   ",
" ..... ",
" .XXX. ",
" .XXX. ",
" .XXX. ",
" ..... ",
" ..... ",
" .GGG. ",
" .GGG. ",
" .GGG. ",
" ..... "};

static	void	*e_sig_pmap[2];		/* R, G */
static	void	*e_sigx_pmap[2];
static const char *e_sig_xpm[] = {
"9 7 3 1",
"       c #FFFFFFFFFFFF",
".      c #000000000000",
NULL, /*"X      c #0000FFFFFFFF",*/
"         ",
"         ",
".    ... ",
".   .GGG.",
".....GGG.",
".   .GGG.",
".    ... "};
static const char *e_sigx_xpm[] = {
"9 7 3 1",
"       c #FFFFFFFFFFFF",
".      c #000000000000",
NULL, /*"X      c #0000FFFFFFFF",*/
"         ",
"         ",
".   .....",
".   .GGG.",
".....GGG.",
".   .GGG.",
".   ....."};

static	void	*w_sig_pmap[2];		/* R, G */
static	void	*w_sigx_pmap[2];
static const char *w_sig_xpm[] = {
"9 7 3 1",
"       c #FFFFFFFFFFFF",
".      c #000000000000",
NULL, /*"X      c #0000FFFFFFFF",*/
"         ",
"         ",
" ...    .",
".GGG.   .",
".GGG.....",
".GGG.   .",
" ...    ."
};
static const char *w_sigx_xpm[] = {
"9 7 3 1",
"       c #FFFFFFFFFFFF",
".      c #000000000000",
NULL, /*"X      c #0000FFFFFFFF",*/
"         ",
"         ",
".....   .",
".GGG.   .",
".GGG.....",
".GGG.   .",
".....   ."
};

static	void	*e_sig2_pmap[4];		/* RR, GR, GG, GO */
static	void	*e_sig2x_pmap[4];
static const char *e_sig2_xpm[] = {
	"13 7 4 1",
	"       c #FFFFFFFFFFFF",
	".      c #000000000000",
	NULL, /*"G      c #0000FFFFFFFF",*/
	NULL, /*"X      c #0000FFFFFFFF",*/
	"             ",
	"             ",
	".   ...  ... ",
	".  .XXX..GGG.",
	"....XXX..GGG.",
	".  .XXX..GGG.",
	".   ...  ... "};
static const char *e_sig2x_xpm[] = {
	"13 7 4 1",
	"       c #FFFFFFFFFFFF",
	".      c #000000000000",
	NULL, /*"G      c #0000FFFFFFFF",*/
	NULL, /*"X      c #0000FFFFFFFF",*/
	"             ",
	"             ",
	".  ..........",
	".  .XXX..GGG.",
	"....XXX..GGG.",
	".  .XXX..GGG.",
	".  .........."};

static	void	*e_sigP_pmap[4];		/* RR, GR, GG, GO */
static const char *e_sigP_xpm[] = {
	"13 7 4 1",
	"       c #FFFFFFFFFFFF",
	".      c #000000000000",
	NULL, /*"G      c #0000FFFFFFFF",*/
	NULL, /*"X      c #0000FFFFFFFF",*/
	"             ",
	"             ",
        ". ...... ... ",
        ". XXXXX..GGG.",
        "....X.X..GGG.",
        ". ..XXX..GGG.",
        ". ...... ... "};


static	void	*w_sig2_pmap[4];		/* RR, GR, GG, GO */
static	void	*w_sig2x_pmap[4];
static const char *w_sig2_xpm[] = {
	"13 7 4 1",
	"       c #FFFFFFFFFFFF",
	".      c #000000000000",
	NULL, /*"G      c #0000FFFFFFFF",*/
	NULL, /*"X      c #0000FFFFFFFF",*/
	"             ",
	"             ",
	" ...  ...   .",
	".GGG..XXX.  .",
	".GGG..XXX....",
	".GGG..XXX.  .",
	" ...  ...   ."};
static const char *w_sig2x_xpm[] = {
	"13 7 4 1",
	"       c #FFFFFFFFFFFF",
	".      c #000000000000",
	NULL, /*"G      c #0000FFFFFFFF",*/
	NULL, /*"X      c #0000FFFFFFFF",*/
	"             ",
	"             ",
	"..........  .",
	".GGG..XXX.  .",
	".GGG..XXX....",
	".GGG..XXX.  .",
	"..........  ."};

static	void	*w_sigP_pmap[4];		/* RR, GR, GG, GO */
static const char *w_sigP_xpm[] = {
	"13 7 4 1",
	"       c #FFFFFFFFFFFF",
	".      c #000000000000",
	NULL, /*"G      c #0000FFFFFFFF",*/
	NULL, /*"X      c #0000FFFFFFFF",*/
	"             ",
	"             ",
	" ... ...... .",
	".GGG..XXX.. .",
	".GGG..X.X....",
	".GGG..XXXXX .",
	" ... ...... ."};

// static
void	Signal::InitPixmaps()
{
	char	bufffg[64];
	char	buff[64];
	int	fgr, fgg, fgb;
	int	r, g, b;
	const char *green_name = "G      c #0000d8000000";

	getcolor_rgb(fieldcolors[COL_TRACK], &fgr, &fgg, &fgb);
	getcolor_rgb(fieldcolors[COL_BACKGROUND], &r, &g, &b);

	sprintf(bufffg, ".      c #%02x00%02x00%02x00", fgr, fgg, fgb);
	sprintf(buff, "       c #%02x00%02x00%02x00", r, g, b);
	sprintf(buff, "       c lightgray", r, g, b);
        n_sig_xpm[1] = s_sig_xpm[1] = e_sig_xpm[1] = w_sig_xpm[1] = 
	n_sigx_xpm[1] = s_sigx_xpm[1] = e_sigx_xpm[1] = w_sigx_xpm[1] = buff;

        n_sig_xpm[2] = s_sig_xpm[2] = e_sig_xpm[2] = w_sig_xpm[2] =
	n_sigx_xpm[2] = s_sigx_xpm[2] = e_sigx_xpm[2] = w_sigx_xpm[2] = bufffg;
        n_sig_xpm[3] = s_sig_xpm[3] = e_sig_xpm[3] = w_sig_xpm[3] =
	n_sigx_xpm[3] = s_sigx_xpm[3] = e_sigx_xpm[3] = w_sigx_xpm[3] = "G      c red";
        n_sig_pmap[0] = get_pixmap(n_sig_xpm);
        s_sig_pmap[0] = get_pixmap(s_sig_xpm);
        e_sig_pmap[0] = get_pixmap(e_sig_xpm);
        w_sig_pmap[0] = get_pixmap(w_sig_xpm);
        n_sigx_pmap[0] = get_pixmap(n_sigx_xpm);
        s_sigx_pmap[0] = get_pixmap(s_sigx_xpm);
        e_sigx_pmap[0] = get_pixmap(e_sigx_xpm);
        w_sigx_pmap[0] = get_pixmap(w_sigx_xpm);

        n_sig_xpm[3] = s_sig_xpm[3] = e_sig_xpm[3] = w_sig_xpm[3] =
	n_sigx_xpm[3] = s_sigx_xpm[3] = e_sigx_xpm[3] = w_sigx_xpm[3] = green_name;
        n_sig_pmap[1] = get_pixmap(n_sig_xpm);
        s_sig_pmap[1] = get_pixmap(s_sig_xpm);
        e_sig_pmap[1] = get_pixmap(e_sig_xpm);
        w_sig_pmap[1] = get_pixmap(w_sig_xpm);
        n_sigx_pmap[1] = get_pixmap(n_sigx_xpm);
        s_sigx_pmap[1] = get_pixmap(s_sigx_xpm);
        e_sigx_pmap[1] = get_pixmap(e_sigx_xpm);
        w_sigx_pmap[1] = get_pixmap(w_sigx_xpm);

	e_sig2_xpm[1] = w_sig2_xpm[1] =
		n_sig2_xpm[1] = s_sig2_xpm[1] =
		e_sig2x_xpm[1] = w_sig2x_xpm[1] =
		n_sig2x_xpm[1] = s_sig2x_xpm[1] =
		e_sigP_xpm[1] = w_sigP_xpm[1] = buff;
	e_sig2_xpm[2] = w_sig2_xpm[2] =
		n_sig2_xpm[2] = s_sig2_xpm[2] =
		e_sig2x_xpm[2] = w_sig2x_xpm[2] =
		n_sig2x_xpm[2] = s_sig2x_xpm[2] =
		e_sigP_xpm[2] = w_sigP_xpm[2] = bufffg;
	e_sig2_xpm[3] = w_sig2_xpm[3] =
		n_sig2_xpm[3] = s_sig2_xpm[3] =
		e_sig2x_xpm[3] = w_sig2x_xpm[3] =
		n_sig2x_xpm[3] = s_sig2x_xpm[3] =
		e_sigP_xpm[3] = w_sigP_xpm[3] = "G      c red";
	e_sig2_xpm[4] = w_sig2_xpm[4] =
		n_sig2_xpm[4] = s_sig2_xpm[4] =
		e_sig2x_xpm[4] = w_sig2x_xpm[4] =
		n_sig2x_xpm[4] = s_sig2x_xpm[4] = "X      c red";
	e_sigP_xpm[4] = w_sigP_xpm[4] = "X      c gray";
	e_sig2_pmap[0] = get_pixmap(e_sig2_xpm);
	w_sig2_pmap[0] = get_pixmap(w_sig2_xpm);
	e_sigP_pmap[0] = get_pixmap(e_sigP_xpm);
	w_sigP_pmap[0] = get_pixmap(w_sigP_xpm);
	n_sig2_pmap[0] = get_pixmap(n_sig2_xpm);
	s_sig2_pmap[0] = get_pixmap(s_sig2_xpm);
	e_sig2x_pmap[0] = get_pixmap(e_sig2x_xpm);
	w_sig2x_pmap[0] = get_pixmap(w_sig2x_xpm);
	n_sig2x_pmap[0] = get_pixmap(n_sig2x_xpm);
	s_sig2x_pmap[0] = get_pixmap(s_sig2x_xpm);

	e_sig2_xpm[3] = w_sig2_xpm[3] =
	n_sig2_xpm[3] = s_sig2_xpm[3] =
	e_sig2x_xpm[3] = w_sig2x_xpm[3] =
	n_sig2x_xpm[3] = s_sig2x_xpm[3] = green_name;
	e_sigP_xpm[3] = w_sigP_xpm[3] = green_name;
	e_sig2_xpm[4] = w_sig2_xpm[4] =
	n_sig2_xpm[4] = s_sig2_xpm[4] =
	e_sig2x_xpm[4] = w_sig2x_xpm[4] =
	n_sig2x_xpm[4] = s_sig2x_xpm[4] = "X      c red";
	e_sigP_xpm[4] = w_sigP_xpm[4] = "X      c gray";
	e_sig2_pmap[1] = get_pixmap(e_sig2_xpm);
	w_sig2_pmap[1] = get_pixmap(w_sig2_xpm);
	e_sigP_pmap[1] = get_pixmap(e_sigP_xpm);
	w_sigP_pmap[1] = get_pixmap(w_sigP_xpm);
	n_sig2_pmap[1] = get_pixmap(n_sig2_xpm);
	s_sig2_pmap[1] = get_pixmap(s_sig2_xpm);
	e_sig2x_pmap[1] = get_pixmap(e_sig2x_xpm);
	w_sig2x_pmap[1] = get_pixmap(w_sig2x_xpm);
	n_sig2x_pmap[1] = get_pixmap(n_sig2x_xpm);
	s_sig2x_pmap[1] = get_pixmap(s_sig2x_xpm);

	e_sig2_xpm[3] = w_sig2_xpm[3] =
	n_sig2_xpm[3] = s_sig2_xpm[3] =
	e_sig2x_xpm[3] = w_sig2x_xpm[3] =
	n_sig2x_xpm[3] = s_sig2x_xpm[3] = green_name;
	e_sig2_xpm[4] = w_sig2_xpm[4] =
	n_sig2_xpm[4] = s_sig2_xpm[4] =
	e_sig2x_xpm[4] = w_sig2x_xpm[4] =
	n_sig2x_xpm[4] = s_sig2x_xpm[4] = "X      c #0000d8000000";
	e_sigP_xpm[4] = w_sigP_xpm[4] = "X      c white";
	e_sig2_pmap[2] = get_pixmap(e_sig2_xpm);
	w_sig2_pmap[2] = get_pixmap(w_sig2_xpm);
	e_sigP_pmap[2] = get_pixmap(e_sigP_xpm);
	w_sigP_pmap[2] = get_pixmap(w_sigP_xpm);
	n_sig2_pmap[2] = get_pixmap(n_sig2_xpm);
	s_sig2_pmap[2] = get_pixmap(s_sig2_xpm);
	e_sig2x_pmap[2] = get_pixmap(e_sig2x_xpm);
	w_sig2x_pmap[2] = get_pixmap(w_sig2x_xpm);
	n_sig2x_pmap[2] = get_pixmap(n_sig2x_xpm);
	s_sig2x_pmap[2] = get_pixmap(s_sig2x_xpm);

	e_sig2_xpm[3] = w_sig2_xpm[3] =
	n_sig2_xpm[3] = s_sig2_xpm[3] =
	e_sig2x_xpm[3] = w_sig2x_xpm[3] =
	n_sig2x_xpm[3] = s_sig2x_xpm[3] = "G      c red";
	e_sigP_xpm[3] = w_sigP_xpm[3] = "G      c red";
	e_sig2_xpm[4] = w_sig2_xpm[4] =
	n_sig2_xpm[4] = s_sig2_xpm[4] =
	e_sig2x_xpm[4] = w_sig2x_xpm[4] =
	n_sig2x_xpm[4] = s_sig2x_xpm[4] = "X      c orange";
	e_sigP_xpm[4] = w_sigP_xpm[4] = "X      c white";
	e_sig2_pmap[3] = get_pixmap(e_sig2_xpm);
	w_sig2_pmap[3] = get_pixmap(w_sig2_xpm);
	e_sigP_pmap[3] = get_pixmap(e_sigP_xpm);
	w_sigP_pmap[3] = get_pixmap(w_sigP_xpm);
	/*n_sig2_pmap[3] = get_pixmap(n_sig2_xpm);
	s_sig2_pmap[3] = get_pixmap(s_sig2_xpm);*/
	e_sig2x_pmap[3] = get_pixmap(e_sig2x_xpm);
	w_sig2x_pmap[3] = get_pixmap(w_sig2x_xpm);
}

// static
void	Signal::FreePixmaps()
{
	int	i;

	for(i = 0; i < 4; ++i) {
	    delete_pixmap(e_sig2_pmap[i]);
	    delete_pixmap(w_sig2_pmap[i]);
	    delete_pixmap(e_sigP_pmap[i]);
	    delete_pixmap(w_sigP_pmap[i]);
	    delete_pixmap(n_sig2_pmap[i]);
	    delete_pixmap(s_sig2_pmap[i]);
	    delete_pixmap(e_sig2x_pmap[i]);
	    delete_pixmap(w_sig2x_pmap[i]);
	    delete_pixmap(n_sig2x_pmap[i]);
	    delete_pixmap(s_sig2x_pmap[i]);
	}
	for(i = 0; i < 2; ++i) {
	    delete_pixmap(n_sig_pmap[i]);
	    delete_pixmap(s_sig_pmap[i]);
	    delete_pixmap(e_sig_pmap[i]);
	    delete_pixmap(w_sig_pmap[i]);
	    delete_pixmap(n_sigx_pmap[i]);
	    delete_pixmap(s_sigx_pmap[i]);
	    delete_pixmap(e_sigx_pmap[i]);
	    delete_pixmap(w_sigx_pmap[i]);
	}
}

void	Signal::Draw()
{
	grcolor color = color_red;
	int	i;
	void	*p = 0;
	Track	*t = this;

	i = 0;					/* RR */
	if(!t->_interpreterData || !(p = FindIcon())) {
	    if(t->fleeted) {
		if(t->status == ST_GREEN) {
		    if(t->nowfleeted)
			i = 2;			/* GG */
		    else
			i = 1;			/* GR */
		} else if(t->nowfleeted)
		    i = 3;				/* RO */
		switch(t->direction) {
		case W_E:
		    p = signal_traditional ?
			(t->signalx ? e_sig2x_pmap[i] : e_sig2_pmap[i]) : e_sigP_pmap[i];
		    break;
		case E_W:
		    p = signal_traditional ?
			(t->signalx ? w_sig2x_pmap[i] : w_sig2_pmap[i]) : w_sigP_pmap[i];
		    break;
		case N_S:
		    p = t->signalx ? s_sig2x_pmap[i] : s_sig2_pmap[i];
		    break;
		case S_N:
		    p = t->signalx ? n_sig2x_pmap[i] : n_sig2_pmap[i];
		    break;
		}
		if(p)
		    draw_pixmap(t->x, t->y, p);
		if(editing && show_links && t->controls)
		    draw_link(t->x, t->y, t->controls->x, t->controls->y, conf.linkcolor);
		return;
	    }
	    if(t->status == ST_GREEN)
		i = 1;
	    switch(t->direction) {
	    case W_E:
		    p = t->signalx ? e_sigx_pmap[i] : e_sig_pmap[i];
		    break;
	    case E_W:
		    p = t->signalx ? w_sigx_pmap[i] : w_sig_pmap[i];
		    break;
	    case N_S:
		    p = t->signalx ? s_sigx_pmap[i] : s_sig_pmap[i];
		    break;
	    case S_N:
		    p = t->signalx ? n_sigx_pmap[i] : n_sig_pmap[i];
		    break;
	    }
	}
	if(p)
	    draw_pixmap(t->x, t->y, p);
	if(editing && show_links && t->controls)
	    draw_link(t->x, t->y, t->controls->x, t->controls->y, conf.linkcolor);
}



SignalAspect::SignalAspect()
{
	_next = 0;
	_name = 0;
	_action = wxStrdup(wxT("none"));
	memset(_iconN, 0, sizeof(_iconN));
	memset(_iconE, 0, sizeof(_iconE));
	memset(_iconS, 0, sizeof(_iconS));
	memset(_iconW, 0, sizeof(_iconW));
}


SignalAspect::~SignalAspect()
{
	int	    i;

	for(i = 0; i < MAX_FLASHING_ICONS; ++i) {
	    if(_iconN[i])
		free(_iconN[i]);
	    if(_iconE[i])
		free(_iconE[i]);
	    if(_iconS[i])
		free(_iconS[i]);
	    if(_iconW[i])
		free(_iconW[i]);
	}
	if(_name)
	    free(_name);
}


bool	Script::ReadFile()
{
	wxChar	*p;

	if(_text)
	    free(_text);
	_text = 0;
	if(!LoadFile(_path, &_text))
	    return false;

	for(p = _text; *p; ) {
	    if(*p == '\t')
		*p++ = ' ';
	    else if(*p == '\r')
		*p++ = '\n';
	    else if(*p == '#') {	// ignore comments
		while(*p && *p != '\n')
		    *p++ = ' ';
	    }
	    else
		++p;
	}
	return true;
}


void	delete_script_data(Track *t)
{
	if(t->type != TSIGNAL)
	    return;
	if(t->_interpreterData) {
	    SignalInterpreterData *interp = (SignalInterpreterData *)t->_interpreterData;

	    while(interp->_aspects) {
		SignalAspect *asp = interp->_aspects;
		interp->_aspects = asp->_next;
		delete asp;
	    }
	    if(interp->_onAuto)
		delete interp->_onAuto;
	    if(interp->_onCleared)
		delete interp->_onCleared;
	    if(interp->_onClick)
		delete interp->_onClick;
	    if(interp->_onCross)
		delete interp->_onCross;
	    if(interp->_onInit)
		delete interp->_onInit;
	    if(interp->_onUpdate)
		delete interp->_onUpdate;

	    free(t->_interpreterData);
	}
	t->_interpreterData = 0;
}


//
//
//	Parse a Train Dir Script file into
//	an Abstract Syntax Tree
//
//	The tree has only 3 types of nodes:
//	Block: beginning of the tree
//	Statement: for assignments
//	If: with optional else and child statements
//
//



static bool match(wxChar **pp, const wxChar *txt)
{
	wxChar	*p = *pp;

	while(*p == ' ' || *p == '\t')
	    ++p;
	*pp = p;
	if(wxStrncmp(p, txt, wxStrlen(txt)))
	    return false;
	p += wxStrlen(txt);
	while(*p == ' ' || *p == '\t')
	    ++p;
	*pp = p;
	return true;
}


static	wxChar *scan_line(wxChar *src, wxChar *dst)
{
	while(*src && *src != '\n') {
	    if(*src == '\r')
		continue;
	    *dst++ = *src++;
	}
	*dst = 0;
	if(*src) ++src;
	while(*src == ' ' || *src == '\t' || *src == '\n' || *src == '\r')
	    ++src;
	return src;
}


static	wxChar	*next_token(wxChar *p)
{
	while(*p && *p == ' ' || *p == '\t' || *p == '\n' || *p == '\r')
	    ++p;
	return p;
}

static	wxChar	*scan_word(wxChar *p, wxChar *dst)
{
	*dst = 0;
	while(*p && (*p == ' ' || *p == '\t')) ++p;
	if(*p >= 'a' && *p <= 'z' || *p >= 'A' && *p <= 'Z' || *p == '_') {
	    while(*p >= 'a' && *p <= 'z' ||
		    *p >= 'A' && *p <= 'Z' ||
		    *p == '_' ||
		    *p >= '0' && *p <= '9')
		*dst++ = *p++;
	    *dst = 0;
	    while(*p && (*p == ' ' || *p == '\t')) ++p;
	    return p;
	}
	if(*p >= '0' && *p <= '9') {
	    while(*p >= '0' && *p <= '9')
		*dst++ = *p++;
	    *dst = 0;
	    while(*p && (*p == ' ' || *p == '\t')) ++p;
	    return p;
	}
	*dst++ = *p++;
	*dst = 0;
	while(*p && (*p == ' ' || *p == '\t')) ++p;
	return p;

}


ExprNode *ParseToken(wxChar **pp)
{
	wxChar	word[256];
	wxChar	*p = *pp;
	ExprNode *n;

	p = scan_word(*pp, word);
	if(!wxStrcmp(word, wxT("switch"))) {
	    n = new ExprNode(SwitchRef);
	} else if(!wxStrcmp(word, wxT("track"))) {
	    n = new ExprNode(TrackRef);
	} else if(!wxStrcmp(word, wxT("signal"))) {
	    n = new ExprNode(SignalRef);
	} else if(!wxStrcmp(word, wxT("next"))) {
	    n = new ExprNode(NextSignalRef);
	} else if(!wxStrcmp(word, wxT("and"))) {
	    n = new ExprNode(And);
	} else if(!wxStrcmp(word, wxT("or"))) {
	    n = new ExprNode(Or);
	} else if(word[0] == '=') {
	    n = new ExprNode(Equal);
	} else if(word[0] == '!') {
	    n = new ExprNode(NotEqual);
	} else if(word[0] == '>') {
	    n = new ExprNode(Greater);
	} else if(word[0] == '<') {
	    n = new ExprNode(Less);
	} else if(word[0] == '.') {
	    n = new ExprNode(Dot);
	} else if(isdigit(word[0])) {
	    n = new ExprNode(Number);
	    n->_val = wxStrtol(word, 0, 0);
	} else if(wxIsalnum(word[0])) {
	    n = new ExprNode(String);
	    n->_txt = wxStrdup(word);
	} else
	    return 0;
	*pp = p;
	return n;
}


ExprNode *ParseExpr(wxChar *p)
{
	ExprNode *n, *n1 = 0, *n2 = 0;
	ExprNode *root = 0;
	wxChar	word[256];

	while(*p) {
	    n = ParseToken(&p);
	    if(!n)
		break;
	    switch(n->_op) {
	    case TrackRef:
	    case SwitchRef:
	    case SignalRef:
		p = scan_word(p, word);
		if(word[0] != '(') {
		    // error: expected '('
		    return 0;
		}
		n1 = ParseToken(&p);
		if(n1->_op != Number) {
		    // error: expected number
		    return 0;
		}
		p = scan_word(p, word);
		if(word[0] != ',') {
		    // error: expected ','
		    return 0;
		}
		n2 = ParseToken(&p);
		if(n2->_op != Number) {
		    // error: expected number
		    return 0;
		}
		n->_x = n1->_val;
		n->_y = n2->_val;
		delete n1;
		delete n2;
		if(!root)
		    root = n;
		p = scan_word(p, word);
		if(word[0] != ')') {
		    // error: expected '('
		    return 0;
		}
		break;

	    case NextSignalRef:
		if(!root)
		    root = n;
		break;

	    case Dot:
		if(!root) {
		    // error: missing left reference;
		    n2 = ParseToken(&p);
		    if(!n2)
			return 0;
		    goto our; // return 0;
		}
		switch(root->_op) {
		case TrackRef:
		case SwitchRef:
		case SignalRef:
		case NextSignalRef:
		case Dot:
		    break;

		default:
		    // error: invalid '.' for left expression
		    return 0;
		}
nxt:
		n2 = ParseToken(&p);
		if(!n2)
		    return 0;
		if(n2->_op == NextSignalRef) {
		    n2->_op = Dot;
		    n2->_left = root;
		    root = n2;
		    continue; // goto nxt;
		}
		if(n2->_op != String) {
		    // error: right of '.' must be a name
		    return 0;
		}
our:		n2->_op = Dot;
		n2->_left = root;
		root = n2;
		break;

	    case Equal:
	    case NotEqual:
	    case Greater:
	    case Less:
		if(!root) {
		    // error: missing left expression
		    return 0;
		}
		n2 = ParseToken(&p);
		if(!n2)
		    return 0;
		n->_left = root;
		n->_right = n2;
		root = n;
		break;

	    case And:
	    case Or:
		if(!root) {
		    // error: missing left expression
		    return 0;
		}
		n2 = ParseExpr(p);	    // recurse!
		if(!n2)
		    return 0;
		n->_left = root;
		n->_right = n2;
		root = n;
		return root;

	    }
	}
	return root;
}


ExprNode::~ExprNode()
{
	if(_op == String && _txt)
	    free(_txt);
	_txt = 0;
	if(_left)
	    delete _left;
	if(_right)
	    delete _right;
}


void	Signal::ParseAspect(wxChar **pp)
{
	wxChar	line[256];
	wxChar	*p = *pp;
	wxChar	**dst;
	SignalAspect *asp = new SignalAspect();
	SignalInterpreterData *interp = (SignalInterpreterData *)_interpreterData;

	p = scan_line(p, line);
	if(line[0])
	    asp->_name = wxStrdup(line);
	do {
	    dst = 0;
	    if(match(&p, wxT("IconN:")))
		dst = &asp->_iconN[0];
	    else if(match(&p, wxT("IconE:")))
		dst = &asp->_iconE[0];
	    else if(match(&p, wxT("IconS:")))
		dst = &asp->_iconS[0];
	    else if(match(&p, wxT("IconW:")))
		dst = &asp->_iconW[0];
	    if(dst) {
		p = scan_line(p, line);
		if(line[0]) {
		    if(wxStrchr(line, ' ')) {
			this->_isFlashing = true;
			int	nxt = 0;
			wxChar	*p1, *pp;

			pp = line;
			do {
			    for(p1 = pp; *pp && *pp != ' '; ++pp);
			    if(p1 != pp) {
				int oc = *pp;
				*pp = 0;
				*dst++ = wxStrdup(p1);
				*pp = oc;
				while(*pp == ' ') ++pp;
				if(++nxt >= MAX_FLASHING_ICONS)
				    break;
			    }
			} while(*pp);
		    } else
			*dst = wxStrdup(line);
		}
		continue;
	    }
	    if(match(&p, wxT("Action:"))) {
		p = scan_line(p, line);
		if(!line[0])
		    continue;
		if(asp->_action)
		    free(asp->_action);
		asp->_action = wxStrdup(line);
		continue;
	    }
	    break;
	    // unknown. Should we give an error?
	} while(1);
	asp->_next = interp->_aspects;
	interp->_aspects = asp;
	*pp = p;
}


SignalStatement *AddStatementToBlock(SignalStatement *block)
{
	SignalStatement *stmt;

	stmt = new SignalStatement();
	if(block->_isElse) {
	    if(!block->_elseChild)
		block->_elseChild = stmt;
	    else
		block->_lastElseChild->_next = stmt;
	    block->_lastElseChild = stmt;
	} else {
	    if(!block->_child)
		block->_child = stmt;
	    else
		block->_lastChild->_next = stmt;
	    block->_lastChild = stmt;
	}
	stmt->_parent = block;
	return stmt;
}


SignalStatement *Signal::ParseStatements(wxChar **pp)
{
	wxChar	line[256];
	wxChar	*p = *pp;
	SignalStatement *stmt;

	SignalStatement *block = new SignalStatement();
	block->_type = 'B';

	while(*p) {
	    if(match(&p, wxT("if"))) {
		stmt = AddStatementToBlock(block);
		stmt->_type = 'I';
		block = stmt;			// enter new scope
		p = scan_line(p, line);
		if(line[0])
		    stmt->_expr = ParseExpr(line);
	    } else if(match(&p, wxT("else"))) {	// exit scope and enter new scope
		do {
		    if(block->_type != 'I')	// else without if?
			return 0;		// error
		    if(!block->_isElse)
			break;
		} while(block = block->_parent);
		if(!block)
		    return 0;
		p = next_token(p);
		block->_isElse = true;
	    } else if(match(&p, wxT("end"))) {	// exit scope
		if(!block->_parent)
		    break;
		block = block->_parent;
		p = next_token(p);
	    } else if(match(&p, wxT("return"))) {	// return from function
		stmt = AddStatementToBlock(block);
		stmt->_type = 'R';		
		p = next_token(p);
	    } else {
		p = scan_line(p, line);
		if(!line[0])
		    continue;
		stmt = AddStatementToBlock(block);
		stmt->_type = 'E';
		stmt->_expr = ParseExpr(line); // wxStrdup(line);
	    }
	}
	*pp = p;
	return block;
}


void	Signal::ParseProgram()
{
	wxChar	*p;

	if(!this->stateProgram || !*this->stateProgram)
	    return;
	Script	*s = find_script(this->stateProgram);
	SignalInterpreterData *interp;
	if(!s) {
	    s = new_script(this->stateProgram);
	    // return;
	}
	if(!s->ReadFile())
	    return;

	if(!_interpreterData)
	    _interpreterData = calloc(sizeof(SignalInterpreterData), 1);

	interp = (SignalInterpreterData *)_interpreterData;
	p = s->_text;
	// todo: seek to Signal: #name
	while(*p) {
	    wxChar	*p1 = p;
	    while(*p1 == ' ' || *p1 == '\t' || *p1 == '\r' || *p1 == '\n')
		++p1;
	    p = p1;
	    if(match(&p, wxT("Aspect:"))) {
		p1 = p;
		ParseAspect(&p);
	    } else if(match(&p, wxT("OnClick:"))) {
		p1 = p;
		interp->_onClick = ParseStatements(&p);
	    } else if(match(&p, wxT("OnCleared:"))) {
		p = next_token(p);
		p1 = p;
		interp->_onCleared = ParseStatements(&p);
	    } else if(match(&p, wxT("OnInit:"))) {
		p = next_token(p);
		p1 = p;
		interp->_onInit = ParseStatements(&p);
	    } else if(match(&p, wxT("OnUpdate:"))) {
		p = next_token(p);
		p1 = p;
		interp->_onUpdate = ParseStatements(&p);
	    } else if(match(&p, wxT("OnAuto:"))) {
		p = next_token(p);
		p1 = p;
		interp->_onAuto = ParseStatements(&p);
	    } else if(match(&p, wxT("OnCross:"))) {
		p = next_token(p);
		p1 = p;
		interp->_onCross = ParseStatements(&p);
	    }
	    if(p1 == p)	    // error! couldn't parse token
		break;
	}
}



//
//	Execution of the Abstract Syntax Tree
//

static	wxChar	expr_buff[1024];

void	Trace(const wxChar *msg)
{
	if(!trace_script)
	    return;
	traindir->AddAlert(msg);
}


Signal	*SignalInterpreterData::GetNextSignal(Signal *sig)
{
	Track	*t;
	trkdir	dir;
	Vector	*path;

	if(!sig->controls)
	    return 0;

	path = findPath(sig->controls, sig->direction);
	t = (Track *)Vector_elementAt(path, path->size - 1);
	dir = (trkdir)Vector_flagAt(path, path->size - 1);
	Vector_delete(path);
	Track *t1;
	if(!(t1 = findNextTrack(dir, t->x, t->y))) {
	    wxSnprintf(expr_buff + wxStrlen(expr_buff), sizeof(expr_buff)/sizeof(wxChar) - wxStrlen(expr_buff), wxT("no next track after (%d,%d)"), t->x, t->y);
	    return 0;
	}
	t = t1;
	sig = (Signal *)((dir == W_E || dir == S_N) ?
			    t->esignal : t->wsignal);
	if(!sig) {
	    wxSnprintf(expr_buff + wxStrlen(expr_buff), sizeof(expr_buff)/sizeof(wxChar) - wxStrlen(expr_buff), wxT("no signal after (%d,%d)"), t->x, t->y);
	    return 0;
	}
	wxSnprintf(expr_buff + wxStrlen(expr_buff), sizeof(expr_buff)/sizeof(wxChar) - wxStrlen(expr_buff), wxT("(%d,%d)"), sig->x, sig->y);
	return sig;
}

bool	SignalInterpreterData::GetNextPath(Signal *s, Vector **ppath)
{
	*ppath = 0;
	if(s->fixedred) {
	    wxStrcat(expr_buff, wxT("signal is always red"));
	    return true;
	}
	*ppath = findPath(s->controls, s->direction);
	if(!*ppath) {
	    wxStrcat(expr_buff, wxT("no valid path"));
	    return false;
	}
	if(_mustBeClearPath) {
//	    if(!s->IsClear()) { // t->status == ST_GREEN) {
//		Vector_delete(path);
//		return true;
//	    }
	    if(pathIsBusy(NULL, *ppath, s->direction)) {
		Vector_delete(*ppath);
		*ppath = 0;
		wxStrcat(expr_buff, wxT("path is busy"));
		return true;
	    }
	}
	return true;
}

bool	SignalInterpreterData::Evaluate(ExprNode *n, ExprValue& result)
{
	Track	*t;
	ExprValue left(None);
	ExprValue right(None);
	const wxChar	*prop;
	int	val;
	int	i;
	Vector	*path;
	Signal	*sig;
	bool	res;

	if(!n)
	    return false;
        switch(n->_op) {

	case SwitchRef:
	case TrackRef:
	case SignalRef:
	// case TriggerRef:
	// case ItineraryRef:
	// case TrainRef:
	
	    wxSnprintf(expr_buff + wxStrlen(expr_buff), sizeof(expr_buff)/sizeof(wxChar) - wxStrlen(expr_buff), wxT("(%d,%d)"), n->_x, n->_y);
	    t = find_track(layout, n->_x, n->_y);
	    if(!t) {
		wxStrcat(expr_buff, wxT("=no track"));
		return false;
	    }
	    result._op = n->_op;
	    result._track = t;
	    return true;

	case NextSignalRef:

	    sig = GetNextSignal(_signal);
	    if(!sig)
		return false;
	    result._op = Addr;
	    result._track = sig;
	    return true;

	case Dot:
	    
	    if(!(n->_left))
		t = this->_signal;
	    else if(n->_left && n->_left->_op == Dot) {
		bool oldForAddr = _forAddr;
		_forAddr = true;
		if(!Evaluate(n->_left, result)) {
		    _forAddr = oldForAddr;
		    return false;
		}
		_forAddr = oldForAddr;
		if(result._op != Addr || result._track->type != TSIGNAL)
		    return false;
		t = GetNextSignal((Signal *)result._track);
	    } else {
		if(!Evaluate(n->_left, result))
		    return false;
		if(result._op != Addr || result._track->type != TSIGNAL)
		    return false;
		t = result._track;
	    }
	    if(!t) {
		wxStrcat(expr_buff, wxT("no current track for '.'"));
		return false;
	    }
	    result._op = Addr;
	    result._track = t;
	    result._txt = n->_txt;
	    wxSnprintf(expr_buff + wxStrlen(expr_buff), sizeof(expr_buff)/sizeof(wxChar) - wxStrlen(expr_buff), wxT("(%d,%d)."), t->x, t->y);
	    if(_forAddr)
		return true;

	    switch(result._op) {
	    
	    case Addr:
	    case SignalRef:
	
		prop = n->_txt;
		wxStrcat(expr_buff, prop);
		if(!wxStrcmp(prop, wxT("aspect"))) {
		    Signal  *s = (Signal *)t;

		    result._op = String;
		    result._txt = s->GetAspect();
		    wxSnprintf(expr_buff + wxStrlen(expr_buff), sizeof(expr_buff)/sizeof(wxChar) - wxStrlen(expr_buff), wxT("{%s}"), result._txt);
		    return true;
		}
		if(!wxStrcmp(prop, wxT("auto"))) {
		    Signal  *s = (Signal *)t;

		    result._op = Number;
		    result._val = s->fleeted;
		    wxSnprintf(expr_buff + wxStrlen(expr_buff), sizeof(expr_buff)/sizeof(wxChar) - wxStrlen(expr_buff), wxT("{%d}"), result._val);
		    return true;
		}
		if(!wxStrcmp(prop, wxT("enabled"))) {
		    Signal  *s = (Signal *)t;

		    result._op = Number;
		    result._val = s->fleeted && s->nowfleeted;
		    wxSnprintf(expr_buff + wxStrlen(expr_buff), sizeof(expr_buff)/sizeof(wxChar) - wxStrlen(expr_buff), wxT("{%d}"), result._val);
		    return true;
		}

		result._op = Number;
		result._val = 0;
		if(!wxStrcmp(prop, wxT("switchThrown"))) {
		    Signal *s = (Signal *)t;

		    res = GetNextPath(s, &path);
		    if(!path)
			return res;

		    for(i = 0; i < path->size; ++i) {
			Track	*trk = (Track *)Vector_elementAt(path, i);

			if(trk->type != SWITCH)
			    continue;
			if(trk->switched) {
			    result._val = 1;
			    break;
			}
		    }
		    wxSnprintf(expr_buff + wxStrlen(expr_buff), sizeof(expr_buff)/sizeof(wxChar) - wxStrlen(expr_buff), wxT("{%s}"), result._val ? wxT("switchThrown") : wxT("switchNotThrown"));
		    Vector_delete(path);
		    return true;
		}
		if(!wxStrcmp(prop, wxT("nextLimit"))) {
		    Signal *s = (Signal *)t;

		    res = GetNextPath(s, &path);
		    if(!path)
			return res;

		    int	    j;
		    int	    lowSpeed = 1000;
		    
		    for(i = 0; i < path->size; ++i) {
			Track	*trk = (Track *)Vector_elementAt(path, i);

			for(j = 0; j < NTTYPES; ++j)
			    if(trk->speed[j] && trk->speed[j] < lowSpeed)
				lowSpeed = trk->speed[j];
		    }
		    result._val = lowSpeed;
		    Vector_delete(path);
		    wxSnprintf(expr_buff + wxStrlen(expr_buff), sizeof(expr_buff)/sizeof(wxChar) - wxStrlen(expr_buff), wxT("{%d}"), lowSpeed);
		    return true;
		}
		if(!wxStrcmp(prop, wxT("nextLength"))) {
		    Signal *s = (Signal *)t;

		    res = GetNextPath(s, &path);
		    if(!path)
			return res;

		    int	    length = 0;
		    
		    for(i = 0; i < path->size; ++i) {
			Track	*trk = (Track *)Vector_elementAt(path, i);
			length += trk->length;
		    }
		    result._val = length;
		    Vector_delete(path);
		    wxSnprintf(expr_buff + wxStrlen(expr_buff), sizeof(expr_buff)/sizeof(wxChar) - wxStrlen(expr_buff), wxT("{%d}"), length);
		    return true;
		}
		if(!wxStrcmp(prop, wxT("nextApproach"))) {
		    Signal *s = (Signal *)t;

		    res = GetNextPath(s, &path);
		    if(!path)
			return res;

		    for(i = 0; i < path->size; ++i) {
			Track	*trk = (Track *)Vector_elementAt(path, i);
			trkdir	dir = (trkdir)Vector_flagAt(path, i);
			if(dir == E_W || dir == N_S) {
			    if(trk->wsignal) {
				Signal *sig = (Signal *)trk->wsignal;
				if(!sig->IsShuntingSignal() && sig->IsApproach()) {
				    result._op = Addr;
				    result._track = sig;
				    break;
				}
			    }
			} else if(trk->esignal) {
			    Signal *sig = (Signal *)trk->esignal;
			    if(!sig->IsShuntingSignal() && sig->IsApproach()) {
				result._op = Addr;
				result._track = sig;
				break;
			    }
			}
		    }
		    Vector_delete(path);
		    wxSnprintf(expr_buff + wxStrlen(expr_buff), sizeof(expr_buff)/sizeof(wxChar) - wxStrlen(expr_buff), wxT(" found"));
		    if(i >= path->size)
			return false;
		    return true;
		}
		if(!wxStrcmp(prop, wxT("nextStation"))) {
		    Signal *s = (Signal *)t;

		    result._op = String;
		    result._txt = wxT("");

		    res = GetNextPath(s, &path);
		    if(!path)
			return res;

		    for(i = 0; i < path->size; ++i) {
			Track	*trk = (Track *)Vector_elementAt(path, i);

			if(!trk->isstation)
			    continue;
			result._txt = trk->station;
			break;
		    }
		    Vector_delete(path);
		    wxSnprintf(expr_buff + wxStrlen(expr_buff), sizeof(expr_buff)/sizeof(wxChar) - wxStrlen(expr_buff), wxT("{%s}"), result._txt);
		    return true;
		}

	    case TrackRef:

		return false;

	    case SwitchRef:
		
		prop = n->_txt;
		if(!wxStrcmp(prop, wxT("thrown"))) {
		    result._op = Number;
		    result._val = t->switched;
		    return true;
		}
	    }
	    return false;

	case String:

	    result._op = n->_op;
	    result._txt = n->_txt;
	    wxStrcat(expr_buff, result._txt);
	    return true;

	case Number:

	    result._op = n->_op;
	    result._txt = n->_txt;
	    result._val = n->_val;
	    wxSnprintf(expr_buff + wxStrlen(expr_buff), sizeof(expr_buff)/sizeof(wxChar) - wxStrlen(expr_buff), wxT("%d"), result._val);
	    return true;

	case Equal:
	    
	    result._op = Number;
	    result._val = 0;
	    if(_forCond) {

		// conditionals return false in case of expression error

		if(!Evaluate(n->_left, left))
		    return true;
		wxStrcat(expr_buff, wxT(" = "));
		if(!Evaluate(n->_right, right))
		    return true;

		val = 0;
		if(left._op == right._op) {
		    switch(left._op) {
		    case String:
			val = !wxStrcmp(left._txt, right._txt);
			break;

		    case Number:
			val = left._val == right._val;
		    }
		}
		result._op = Number;
		result._val = val;
		wxSnprintf(expr_buff + wxStrlen(expr_buff), sizeof(expr_buff)/sizeof(wxChar) - wxStrlen(expr_buff), wxT("{%d}"), val);
		return true;
	    }
	    _forAddr = true;
	    if(!Evaluate(n->_left, result))
		return false;
	    _forAddr = false;
	    wxStrcat(expr_buff, wxT(" = "));
	    if(!Evaluate(n->_right, right))
		return false;
	    switch(result._op) {
	    
	    case TrackRef:
	    case SignalRef:
		
		return false;

	    case SwitchRef:
	    case Addr:
		
		prop = result._txt;
		if(!wxStrcmp(prop, wxT("thrown"))) {
		    // call t->Throw(right._val);
		} else if(!wxStrcmp(prop, wxT("aspect"))) {
		    Signal  *s = (Signal *)result._track;
		    
		    s->SetAspect(right._txt);
		} else if(!wxStrcmp(prop, wxT("auto"))) {
		    Signal  *s = (Signal *)result._track;
		    
		    s->fleeted = right._val;
		} else if(!wxStrcmp(prop, wxT("enabled"))) {
		    Signal  *s = (Signal *)result._track;
		    
		    s->nowfleeted = right._val;
		} else if(!wxStrcmp(prop, wxT("fleeted"))) {
		    Signal  *s = (Signal *)result._track;
		    
		    s->fleeted = right._val;
		    s->nowfleeted = right._val;
		} else if(!wxStrcmp(prop, wxT("shunting"))) {
		    Signal  *s = (Signal *)result._track;
		    
		    s->_isShuntingSignal = right._val != 0;
		}
	    }
	    return true;

	case NotEqual:

	    result._op = Number;
	    result._val = 0;
	    if(!Evaluate(n->_left, left)) {
		result._val = 1;	// invalid expressions never match
		return true;
	    }
	    wxStrcat(expr_buff, wxT(" != "));
	    if(!Evaluate(n->_right, right))
		return true;
	    val = 0;
	    if(left._op == right._op) {
		switch(result._op) {
		case String:
		    val = wxStrcmp(left._txt, right._txt) != 0;
		    break;

		case Number:
		    val = left._val != right._val;
		}
	    }
	    result._val = val;
	    wxSnprintf(expr_buff + wxStrlen(expr_buff), sizeof(expr_buff)/sizeof(wxChar) - wxStrlen(expr_buff), wxT("{%d}"), val);
	    return true;

	case Greater:

	    // conditionals return false in case of expression error

	    result._op = Number;
	    result._val = 0;
	    if(!Evaluate(n->_left, left))
		return true;
	    wxStrcat(expr_buff, wxT(" > "));
	    if(!Evaluate(n->_right, right))
		return true;
	    val = 0;
	    if(left._op == right._op) {
		switch(left._op) {
		case String:
		    val = wxStrcmp(left._txt, right._txt) > 0;
		    break;

		case Number:
		    val = left._val > right._val;
		}
	    }
	    result._val = val;
	    wxSnprintf(expr_buff + wxStrlen(expr_buff), sizeof(expr_buff)/sizeof(wxChar) - wxStrlen(expr_buff), wxT("{%d}"), val);
	    return true;

	case Less:

	    // conditionals return false in case of expression error

	    result._op = Number;
	    result._val = 0;
	    if(!Evaluate(n->_left, left))
		return true;
	    wxStrcat(expr_buff, wxT(" < "));
	    if(!Evaluate(n->_right, right))
		return true;
	    val = 0;
	    if(left._op == right._op) {
		switch(left._op) {
		case String:
		    val = wxStrcmp(left._txt, right._txt) < 0;
		    break;

		case Number:
		    val = left._val < right._val;
		}
	    }
	    result._val = val;
	    wxSnprintf(expr_buff + wxStrlen(expr_buff), sizeof(expr_buff)/sizeof(wxChar) - wxStrlen(expr_buff), wxT("{%d}"), val);
	    return true;

	case Or:

	    result._op = Number;
	    result._val = 0;
	    if(Evaluate(n->_left, right) && right._op == Number && right._val != 0) {
		result._val = 1;
		return true;
	    }
	    wxStrcat(expr_buff, wxT(" or "));
	    // note: invalid expressions evaluate to false (0)
	    if(Evaluate(n->_right, right) && right._op == Number && right._val != 0)
		result._val = 1;
	    else
		result._val = 0;
	    return true;

	case And:

	    result._op = Number;
	    result._val = 0;
	    if(Evaluate(n->_left, right) && right._op == Number && right._val == 0)
		return true;
	    wxStrcat(expr_buff, wxT(" and "));
	    // note: invalid expressions evaluate to false (0)
	    if(Evaluate(n->_right, right) && right._op == Number && right._val == 0)
		return true;
	    result._val = 1;
	    return true;

	default:

	    return false;
	}
	return false;
}


void	SignalInterpreterData::Execute(SignalStatement *stmt)
{
	ExprValue   result(None);
	bool	    valid;

	while(stmt) {
	    _scopes[_stackPtr] = stmt;

	    switch(stmt->_type) {
	    case 'B':
		++_stackPtr;
		stmt = stmt->_child;
		continue;

	    case 'I':
		wxStrcpy(expr_buff, wxT("if "));
		result._op = None;
		_forCond = true;
		valid = Evaluate(stmt->_expr, result);
		if(!valid) {
		    if(stmt->_elseChild) {
			wxStrcat(expr_buff, wxT(" -> else"));
			Trace(expr_buff);
			++_stackPtr;
			stmt = stmt->_elseChild;
			continue;
		    }
		    wxStrcat(expr_buff, wxT(" -> false"));
		} else if(result._op == Number) {
		    if(result._val) {
			wxStrcat(expr_buff, wxT(" -> true"));
			Trace(expr_buff);
			++_stackPtr;
			stmt = stmt->_child;
			continue;
		    }
		    if(stmt->_elseChild) {
			wxStrcat(expr_buff, wxT(" -> else"));
			Trace(expr_buff);
			++_stackPtr;
			stmt = stmt->_elseChild;
			continue;
		    }
		    wxStrcat(expr_buff, wxT(" -> false"));
		} else
		    wxStrcat(expr_buff, wxT(" * Result not a number"));
		Trace(expr_buff);
		break;

	    case 'R':
		return;

	    case 'E':
		result._op = None;
		_forCond = false;
		expr_buff[0] = 0;
		Evaluate(stmt->_expr, result);
		Trace(expr_buff);
	    }
	    while(!stmt->_next) {
		if(!_stackPtr)
		    return;
		stmt = _scopes[--_stackPtr];
	    }
	    stmt = stmt->_next;
	}
}



void	*Signal::FindIcon()
{
	SignalInterpreterData *interp = (SignalInterpreterData *)_interpreterData;
	SignalAspect	*asp = interp->_aspects;
	wxChar		**p = 0;
	int		ix;
	const wxChar		*curState;

	if(this->_currentState)
	    curState = this->_currentState;
	else if(this->status == ST_GREEN)
	    curState = wxT("green");
	else
	    curState = wxT("red");

	while(asp) {
	    if(!wxStricmp(asp->_name, curState))
		break;
	    asp = asp->_next;
	}
	if(!asp)
	    return 0;
	switch(this->direction) {
	case W_E:
		p = asp->_iconE;
		break;

	case E_W:
		p = asp->_iconW;
		break;

	case N_S:
		p = asp->_iconS;
		break;

	case S_N:
		p = asp->_iconN;
		break;
	}
	if(!p || !*p)
	    return 0;
	if(_isFlashing) {
	    if(!p[_nextFlashingIcon])
		_nextFlashingIcon = 0;
	    p = &p[_nextFlashingIcon];
	}
	if((ix = get_pixmap_index(*p)) < 0)
	    return 0;
	return pixmaps[ix].pixels;
}




bool	Signal::IsClear()
{
	if(this->_currentState)
	    return wxStrcmp(GetAction(), wxT("stop")) != 0;	// !Rask
	return this->status == ST_GREEN;
}


void	Signal::OnClear()
{
	signals_changed = 1;
	if(_interpreterData) {
	    SignalInterpreterData *interp = (SignalInterpreterData *)_interpreterData;
	    if(interp->_onCleared) {
		interp->_signal = this;
		interp->_stackPtr = 0;
		interp->_mustBeClearPath = true;
		wxSnprintf(expr_buff, sizeof(expr_buff)/sizeof(wxChar), wxT("%s::OnClear(%d,%d)"), this->stateProgram, this->x, this->y);
		Trace(expr_buff);
		interp->Execute(interp->_onCleared);
		return;
	    }
	}

	this->status = ST_GREEN;
	_currentState = wxT("green");
	_nextFlashingIcon = 0;	    // in case new aspect is not flashing
}


void	Signal::OnUnclear()
{
	signals_changed = 1;
	if(this->_interpreterData) {
	    SignalInterpreterData *interp = (SignalInterpreterData *)_interpreterData;
	    if(interp->_onClick) {
		interp->_signal = this;
		interp->_stackPtr = 0;
		interp->_mustBeClearPath = false;
		wxSnprintf(expr_buff, sizeof(expr_buff)/sizeof(wxChar), wxT("%s::OnUnclear(%d,%d)"), this->stateProgram, this->x, this->y);
		Trace(expr_buff);
		interp->Execute(interp->_onClick);
		return;
	    }
	}

	this->status = ST_RED;
	_currentState = wxT("red");
	_nextFlashingIcon = 0;	    // in case new aspect is not flashing
}



void	Signal::OnCross()
{
//	if(this->aspect_changed)
//	    return;

	SignalInterpreterData *interp = (SignalInterpreterData *)_interpreterData;
	if(!_interpreterData || !interp->_onCross) {
	    signals_changed = 1;
	    this->status = ST_RED;
	    SetAspect(wxT("red"));
	    _nextFlashingIcon = 0;	    // in case new aspect is not flashing
	    return;
	}

	interp->_signal = this;
	interp->_stackPtr = 0;
	interp->_mustBeClearPath = false;   // ???
	wxSnprintf(expr_buff, sizeof(expr_buff)/sizeof(wxChar), wxT("%s::OnCross(%d,%d)"), this->stateProgram, this->x, this->y);
	Trace(expr_buff);
	interp->Execute(interp->_onCross);
}


void	Signal::OnUnlock()
{
	signals_changed = 1;
	this->status = ST_GREEN;
}


void	Signal::OnUnfleet()
{
	signals_changed = 1;
	this->status = ST_GREEN;
}

void	Signal::OnUpdate()
{
	if(this->aspect_changed)
	    return;

	if(!this->_interpreterData)
	    return;

	SignalInterpreterData *interp = (SignalInterpreterData *)_interpreterData;
	if(!interp->_onUpdate)
	    return;

	interp->_signal = this;
	interp->_stackPtr = 0;
	interp->_mustBeClearPath = false;
	wxSnprintf(expr_buff, sizeof(expr_buff)/sizeof(wxChar), wxT("%s::OnUpdate(%d,%d)"), this->stateProgram, this->x, this->y);
	Trace(expr_buff);
	interp->Execute(interp->_onUpdate);
}

void	Signal::OnInit()
{
	SignalInterpreterData *interp = (SignalInterpreterData *)_interpreterData;
	if(!interp)
	    return;
	if(!interp->_onInit)
	    return;
	interp->_signal = this;
	interp->_stackPtr = 0;
	interp->_mustBeClearPath = true;
	wxSnprintf(expr_buff, sizeof(expr_buff)/sizeof(wxChar), wxT("%s::OnInit(%d,%d)"), this->stateProgram, this->x, this->y);
	Trace(expr_buff);
	interp->Execute(interp->_onInit);
	return;
}

void	Signal::OnAuto()
{
	if(!this->_interpreterData)
	    return;

	SignalInterpreterData *interp = (SignalInterpreterData *)_interpreterData;
	if(!interp->_onAuto)
	    return;

	interp->_signal = this;
	interp->_stackPtr = 0;
	interp->_mustBeClearPath = true;
	wxSnprintf(expr_buff, sizeof(expr_buff)/sizeof(wxChar), wxT("%s::OnAuto(%d,%d)"), this->stateProgram, this->x, this->y);
	Trace(expr_buff);
	interp->Execute(interp->_onAuto);
}


void	Signal::OnFlash()
{
	SignalAspect *asp;
	SignalInterpreterData *interp = (SignalInterpreterData *)_interpreterData;
	wxChar	    **p;

	if(!interp || !_currentState)
	    return;

	for(asp = interp->_aspects; asp; asp = asp->_next)
	    if(!wxStrcmp(_currentState, asp->_name)) {
		int	nxt = this->_nextFlashingIcon + 1;

		if(nxt >= MAX_FLASHING_ICONS)
		    nxt = 0;
		p = 0;
		switch(this->direction) {
		case W_E:
			p = asp->_iconE;
			break;

		case E_W:
			p = asp->_iconW;
			break;

		case N_S:
			p = asp->_iconS;
			break;

		case S_N:
			p = asp->_iconN;
			break;
		}
		if(!p || ! p[nxt])
		    nxt = 0;
		this->_nextFlashingIcon = nxt;
    		change_coord(this->x, this->y);
		break;
	    }
}


void	Signal::SetAspect(const wxChar *aspect)
{
	if(!_currentState || wxStrcmp(_currentState, aspect)) {
	    signals_changed = 1;
    	    change_coord(this->x, this->y);
	    this->aspect_changed = 1;
	}

	_currentState = aspect;
	_nextFlashingIcon = 0;	    // in case new aspect is not flashing
}


const wxChar *Signal::GetAspect()
{
	if(_currentState)
	    return _currentState;
	if(this->status == ST_RED)
	    return wxT("red");
	return wxT("green");
}

const wxChar	*Signal::GetAction()
{
	const wxChar	*name = GetAspect();
	SignalInterpreterData *interp = (SignalInterpreterData *)_interpreterData;
	SignalAspect *asp;

	if(!interp) {
	    if(!wxStrcmp(name, wxT("red")))
		return wxT("stop");
	    return wxT("proceed");
	}
	for(asp = interp->_aspects; asp; asp = asp->_next) {
	    if(!wxStrcmp(name, asp->_name) && asp->_action)
		return asp->_action;
	}
	return wxT("proceed");	    // broken signal? maybe we should stop.
}


bool	Signal::IsApproach()
{
	return wxStrcmp(GetAction(), wxT("none")) == 0 || _isShuntingSignal;
}


bool	Signal::IsShuntingSignal()
{
	return _isShuntingSignal;
}

bool	Signal::GetSpeedLimit(int *limit)
{
	wxChar	buff[256];
	wxChar	*action = (wxChar *)GetAction();
	action = scan_word(action, buff);

	if(!wxStrcmp(buff, wxT("speedLimit"))) {
	    *limit = wxAtoi(action);
	    return true;
	}
	return false;
}

