#ifndef _TSIGNAL_H
#define _TSIGNAL_H

class Signal;

#define	MAX_FLASHING_ICONS  4

class SignalAspect {
public:
	SignalAspect();
	~SignalAspect();

	SignalAspect *_next;
	wxChar	    *_name;
	wxChar	    *_iconN[MAX_FLASHING_ICONS],
		    *_iconE[MAX_FLASHING_ICONS],
		    *_iconS[MAX_FLASHING_ICONS],
		    *_iconW[MAX_FLASHING_ICONS];
	wxChar	    *_action;
};


class SignalAction {
public:
	SignalAction();
	~SignalAction();


	SignalAction *_next;
	wxChar	    *_name;
	wxChar	    *_program;
};

extern	SignalAction	*signalActionList;


class SignalStatement {
public:
	SignalStatement()
	{
	    _next = _child = _lastChild = _parent = 0;
	    _elseChild = _lastElseChild = 0;
	    _isElse = false;
	    _type = 0;
	    _text = 0;	    // todo: remove?
	    _expr = 0;
	}
	
	~SignalStatement()
	{
	    while(_elseChild) {
		SignalStatement *stmt = _elseChild;
		_elseChild = _elseChild->_next;
		delete stmt;
	    }
	    _lastElseChild = 0;
	    while(_child) {
		SignalStatement *stmt = _child;
		_child = _child->_next;
		delete stmt;
	    }
	    _lastChild = 0;
	    if(_expr)
		delete _expr;
	    if(_text)
		free(_text);
	    _text = 0;
	}

	SignalStatement *_next;
	SignalStatement *_child, *_elseChild;
	SignalStatement *_lastChild, *_lastElseChild;
	SignalStatement *_parent;
	int		_type;
	bool		_isElse;
	wxChar		*_text;
	ExprNode	*_expr;
};

// Track::interpreterData for signals

#define	MAXNESTING  10

class SignalInterpreterData {
public:
	SignalAspect *_aspects;	    // list of aspects (states)
	SignalStatement *_onClick;  // list of actions (statements)
	SignalStatement *_onCleared;// list of actions (statements)
	SignalStatement *_onUpdate; // list of actions (statements)
	SignalStatement *_onInit;   // list of actions (statements)
	SignalStatement	*_onCross;  // list of actions (statements)
	SignalStatement *_onAuto;   // list of actions (statements)

	// run-time environment

	void		Execute(SignalStatement *stmt);
	bool		Evaluate(ExprNode *expr, ExprValue& result);
	Signal		*GetNextSignal(Signal *sig);

	bool		GetNextPath(Signal *sig, Vector **ppath);

	bool		_forCond;
	bool		_forAddr;
	bool		_mustBeClearPath;
	int		_stackPtr;
	SignalStatement *_scopes[MAXNESTING];
	Signal		*_signal;
	Signal		*_curSignal;// support for next.next chains
};



class Signal : public Track {
public:

	static void InitPixmaps();

	static void FreePixmaps();

	void	Draw();

	bool	IsClear();	// is signal green?

	bool	IsApproach();	// is signal an approach signal?

	bool	IsShuntingSignal(); // is a shunting signal?

	void	OnClear();	// set signal to green

	void	OnUnclear();	// set signal to red

	void	OnCross();	// set signal to red when a train enters the controlled section

	void	OnUnlock();	// set signal to green after path has become clear

	void	OnUnfleet();	// set fleeted signal to green after path has become clear

	void	OnUpdate();	// some other signal change, see if we need to change, too

	void	OnInit();	// initial setting (when load or restart)

	void	OnFlash();	// display next flashing aspect

	void	OnAuto();	// automatic signal has been enabled/disabled

	void	ParseProgram();	// fill Track::interpreterData

	void	FreeProgram();	// release Track::interpreterData


	void	*FindIcon();	// find icon for current aspect and direction

	void	ParseAspect(wxChar **p);

	void	SetAspect(const wxChar *p); // called at run-time

	const wxChar	*GetAspect();	// called at run-time

	const wxChar	*GetAction();	// get action for current aspect

	bool	GetSpeedLimit(int *limit);  // get speed limit associated
					    // with current aspect

	SignalStatement *ParseStatements(wxChar **p);
};

#endif // _TSIGNAL_H
