/*	loadsave.cpp - Created by Giampiero Caprino

This file is part of Train Director 3

Train Director is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; using exclusively version 2.
It is expressly forbidden the use of higher versions of the GNU
General Public License.

Train Director is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Train Director; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if !defined(__unix__) && !defined(__WXMAC__)
#include <malloc.h>
#else
#define	stricmp strcasecmp
#endif

#include <memory.h>
#include <string.h>
#include "Traindir3.h"

wxChar	*script_text;		// program-oriented control

void	append_to_script(const wxChar *txt)
{
	wxChar	*p;

	if(!script_text)
	    script_text = wxStrdup(txt);
	else {
	    p = (wxChar *)malloc((wxStrlen(script_text) + wxStrlen(txt) + 4) * sizeof(wxChar));
	    wxStrcpy(p, script_text);
	    wxStrcat(p, txt);
	    free(script_text);
	    script_text = p;
	}
}

void	add_to_script(Track *t)
{
	wxChar	buff[512];
	wxChar	*p;
	wxChar	*p1;
	int	i;

	while(t) {
	    if(t->type == TRIGGER && t->station) {
		wxSnprintf(buff, sizeof(buff)/sizeof(wxChar), wxT("On Trigger %d, %d\n"), t->x, t->y);
		append_to_script(buff);
		if((p = wxStrchr(t->station, '{'))) {
		    wxStrcpy(buff, wxT("\tIf Train.Name = "));
		    for(p1 = wxStrchr (buff, '\0'); *++p && *p != '}'; *p1++ = *p);
		    *p1++ = '\n';
		    *p1 = 0;
		    append_to_script(buff);
		    p1 = wxT("\t\t");
		} else
		    p1 = wxT("\t");

		for(i = 0; i < NTTYPES; ++i) {
		    if(t->speed[i] && t->speed[i] != 100) {
			wxSnprintf(buff, sizeof(buff)/sizeof(wxChar), wxT("%sIf Train.Type = %d AND RANDOM < %d%%\n"), p1, i, t->speed[i]);
			append_to_script(buff);
		    }
		}
	    }
	    t = t->next;
	}
}

//	"click x,y",
//	"rclick x,y",
//	"fast",
//	"slow",
//	"shunt train",
//	"reverse train",
//	"traininfo train",
//	"stationinfo train",
//	"accelerate speed train",
//	"decelerate speed train",

enum	TokenId {
    K_ERR,
    K_ON,
    K_IF,
    K_OR,
    K_AND,
    K_TRAIN,
    K_TRIGGER,
    K_TYPE,
    K_RANDOM,
    K_NAME,
    K_CLICK,
    K_RCLICK,
    K_FAST,
    K_SLOW,
    K_SHUNT,
    K_REVERSE,
    K_TRAININFO,
    K_STATIONINFO,
    K_ACCELERATE,
    K_DECELERATE
};

struct Token {
	const wxChar	*txt;
	TokenId	key;
} tokens[] = {
    {	wxT("ON"),	K_ON },
    {	wxT("IF"),	K_IF },
    {	wxT("OR"),	K_OR },
    {	wxT("AND"),	K_AND },
    {	wxT("TRAIN"),  K_TRAIN },
    {	wxT("TYPE"),	  K_TYPE },
    {	wxT("RANDOM"), K_RANDOM },
    {	wxT("NAME"),	  K_NAME },
    {	wxT("CLICK"),  K_CLICK },
    {	wxT("RCLICK"), K_RCLICK },
    {	wxT("FAST"),	  K_FAST },
    {	wxT("SLOW"),	  K_SLOW },
    {	wxT("SHUNT"),  K_SHUNT },
    {	wxT("REVERSE"), K_REVERSE },
    {	wxT("TRAININFO"), K_TRAININFO },
    {	wxT("STATIONINFO"), K_STATIONINFO },
    {	wxT("ACCELERATE"),  K_ACCELERATE },
    {	wxT("DECELERATE"),  K_DECELERATE },
    {	wxT("TRIGGER"),     K_TRIGGER },
    {	0 }
};

struct Statement {
	Statement *next;
	TokenId	token;
	TokenId objtype;
	Statement *expr;
	Statement *left, *right;
	int	x, y;
	int	value;
	wxChar	*values;
};

Statement   *script;

Statement *isToken(const wxChar *p)
{
	int	i;
	Statement *s;

	for(i = 0; tokens[i].txt; ++i) {
	    if(!wxStricmp(tokens[i].txt, p)) {
		s = (Statement *)calloc(sizeof(Statement), 1);
		s->token = tokens[i].key;
		return s;
	    }
	}
	return 0;
}

Statement *free_statement(Statement *parent, Statement *s)
{
	Statement *t;

	if(s == parent) {
	    t = s->next;
	} else {
	    t = 0;
	    while(parent->next) {
		if(parent->next == s) {
		    parent->next = s->next;
		    t = parent;
		    break;
		}
	    }
	}
	while(s->left)
	    s->left = free_statement(s->left, s->left);
	while(s->right)
	    s->right = free_statement(s->right, s->right);
	while(s->expr)
	    s->expr = free_statement(s->expr, s->expr);
	if(s->values)
	    free(s->values);
	free(s);
	return t;
}

void	free_statements()
{
	while(script) {
	    script = free_statement(script, script);
	}
}

bool	parse_script()
{
	free_statements();
	return false;
}

//
//
//
//

Script	*scriptList;


void	free_scripts()
{
	Script	*p;

	while(scriptList) {
	    p = scriptList;
	    scriptList = scriptList->_next;
	    if(p->_path)
		free(p->_path);
	    if(p->_text)
		free(p->_text);
	    free(p);
	}
}

Script	*find_script(const wxChar *path)
{
	Script	*s;
	wxChar	buff[256];
	wxChar	*p;

	wxStrncpy(buff, path, sizeof(buff) / sizeof(wxChar));
	if((p = wxStrchr(buff, '#')))
	    *p = 0;
	for(s = scriptList; s; s = s->_next) {
	    if(!wxStrcmp(s->_path, buff))
		return s;
	}
	return 0;
}

Script	*new_script(const wxChar *path)
{
	Script	*s;
	wxChar	buff[256];
	wxChar	*p;

	wxStrncpy(buff, path, sizeof(buff) / sizeof(wxChar));
	if((p = wxStrchr(buff, '#')))
	    *p = 0;
	s = (Script *)calloc(sizeof(Script), 1);
	s->_next = scriptList;
	scriptList = s;
	s->_path = wxStrdup(buff);
	s->_text = 0;
	return s;
}


//	load_scripts
//	    Collect all script file names from signals
//	    (and eventually itineraries) in the list
//	    scriptList. This allows multiple signals to
//	    share the same script file.

void	load_scripts(Track *trk)
{
	for(; trk; trk = trk->next) {
	    if(trk->type != TSIGNAL)
		continue;
	    if(!trk->stateProgram)
		continue;
//	    new_script(trk->stateProgram);
	    Signal *sig = (Signal *)trk;
	    sig->ParseProgram();
	    sig->OnInit();
	}
}


